/******************************************/
/*                                        */
/*         ---  callpgm.c ---             */
/*                                        */
/*   Copyright (C) 1991 by MarshallSoft   */
/*   Box 4543, Huntsville AL 35815        */
/*                                        */
/******************************************/

#include <stdio.h>
#include <fcntl.h>
#include <dos.h>
#include <sys\types.h>
#include <sys\stat.h>
#include "ccl.h"

#define FALSE 0
#define TRUE !FALSE
#define ONESECOND 30
#define THREESECONDS (3*ONESECOND)
#define TENSECONDS (10*ONESECOND)
#define BYTE unsigned char
#define BUFFERSIZE 1024
#define MAXTRY 5
#define LIMIT 20
#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define BS  0x08
#define CR  0x0d
#define XON 0x11
#define XOFF 0x13
#define NAK 0x15
#define CAN 0x18
#define CTLZ 0x1a
#define ESC 0x1b
#define NORMAL 0x07
#define INVERSE 0x70

/*** Global Variables ***/

int  Port;               /* current COM port [0..3] */
char Filename[15];       /* file name buffer */
char far RxBuf[1024];    /* CCL receive buffer */
char *BaudRate[10] =  {"300","600","1200","2400","4800","9600",
                       "19200","38400","57600","115200"};
char temp[80];           /* temporary string buffer */
char Buffer[1024];       /* disk buffer */

/*** Main ***/

main(argc,argv)
int argc;
char *argv[];
{int DataFlag = FALSE;
 char c;
 int i, k;
 int n, rc;
 int Delta;         /* delta port status */
 int Status;        /* port status */
 int BaudCode;      /* baud rate code ( index into BaudRate[] )
 /* check for right number of arguments */
 if(argc!=3)
     {printf("Usage: 'CALLPGM port baud' -- example 'CALLPGM 1 9600'\n");
      exit(1);
     }
 /* get port # from command line */
 Port = atoi(argv[1]) - 1;
 if((Port<0) || (Port>3))
     {printf("Port must be 1 to 4\n");
      exit(1);
     }
 /* get baud rate from command line */
 BaudCode = BaudMatch(argv[2]);
 if(BaudCode<0)
     {printf("Cannot recognize baud rate = %s\n",argv[2]);
      exit(1);
     }
 /* initialize screen */
 Scroll(0,0,24,79,0,NORMAL);
 /* setup receive buffer */
 rc = SioRxBuf(Port,RxBuf,Size1024);
 if(rc<0)
     {SioError(rc);
      exit(1);
     }
 /* set parms & reset (initialize) COM port */
 SioParms(Port,NoParity,OneStopBit,WordLength8);
 MyStart(Port,BaudCode);
 /* set DTR and RTS */
 SioDTR(Port,SET);
 SioRTS(Port,SET);
 /* wait for DSR */
#if 0
 printf("Waiting for DSR.");
 while( !SioModem(Port,DSR) )
     {
      if(SioKeyPress()||SioBrkKey()) MyExit(0,"Aborted by user");
      putchar('.');
      SioDelay(18);
     }
#endif
 /* initialize screen */
 Scroll(0,0,24,79,0,NORMAL);
 /* display status message */
 sprintf(temp," COM%d %s 'ESC for menu' ",1+Port,BaudRate[BaudCode]);
 for(i=0;i<strlen(temp);i++)
     {/* write char */
      Position(24,50+i);
      AttrWrite(temp[i],INVERSE);
     }
 /* home cursor & clear CCL receive buffer */
 Position(0,0);
 SioRxFlush(Port);
 /* enter terminal loop */
 while(1)
     {/* Control-BREAK ? */
      if(SioBrkKey()) MyExit(0,"User pressed Ctrl-BREAK");
      /* was key pressed ? */
      if(SioKeyPress())
          {i = SioKeyRead();
           if((char)i==ESC)
               {/* process user's request */
                ProcessESC();
                continue;
               }
           else PutChar(Port,i);
          }
      /* was break detected ? */
      rc = SioBrkSig(Port,DETECT);
      if(rc<0) MyExit(rc,"Exit-B");
      if(rc) MyExit(0,"BREAK detected");
      /* any incoming over serial port ? */
      i = GetChar(Port,0);
      if(i>-1)
          {/* good character */
           if(DataFlag&(i<0x20))
               {MyCrtWrite('^');
                MyCrtWrite('@'+i);
               }
           else MyCrtWrite(i&0x00ff);
          }
      /* any change in DCD or DSR ? */
      Delta = SioModem(Port,DeltaDCD|DeltaDSR);
      if(Delta<0) MyExit(Delta,"Exit-D");
      if(Delta)
          {/* display new status */
           Status = SioModem(Port,DCD|DSR);
           if(Status<0) MyExit(Status,"Exit-E");
           if(!Delta&DeltaDCD) MyExit(0,"Dropped DSD");
           if(!Delta&DeltaDSR) MyExit(0,"Dropped DSR");
          }
     } /* end -- key pressed */
}

/*** write character to screen ***/

int MyCrtWrite(ch)
char ch;
{/* write character */
 SioCrtWrite(ch);
 /* scroll all but bottom line */
 if(GetRow()==24)
    {Scroll(0,0,23,79,1,NORMAL);
     Position(23,0);
    }
}

/*** reset COM port ***/

int MyStart(Port,BaudCode)
int Port;
int BaudCode;
{int i, rc;
 /* try up to 5 times to reset COM port */
 for(i=0;i<5;i++)
     {printf("Resetting COM%d at %s baud\n",Port+1,BaudRate[BaudCode]);
      if( (rc = SioReset(Port,BaudCode))==0) return;
      if(rc<0) MyExit(rc,"Exit-F");
      SioDone(Port);
      /* display errors */
      if(rc&OverrunError) puts("Overrun Error");
      if(rc&ParityError)  puts("Parity Error");
      if(rc&FramingError) puts("Framing Error");
      if(rc&BreakDetect)  puts("Break Detect");
      SioDelay(9);
     }
 exit(1);
}

/*** Lookup baud rate in tabele ***/

int BaudMatch(ptr)
char *ptr;
{int i;
 /* find baud rate in table */
 for(i=0;i<10;i++) if(strcmp(BaudRate[i],ptr)==0) return(i);
 return(-1);
}

/*** Process user request ***/

int ProcessESC()
{int i;
 int rc;
 int c1, c2;
 char Answer[2]; /* array for 1 char answer */
 int row, col;
 /* user pressed <ESC> */
 DisplayLine(" Q) Quit S) Send R) Receive: ",Answer,1);
 if(strlen(Answer)) switch(toupper(Answer[0]))
     {
      case 'Q':
          MyExit(0,"User pressed ESC");
          break;
      case 'R':
          DisplayLine(" Enter filename:",Filename,15);
          if(strlen(Filename)==0) break;
          RxAscii(Port,Filename,Buffer,1024,1024);
          break;
      case 'S':
          DisplayLine(" Enter filename:",Filename,15);
          if(strlen(Filename)==0) break;
          TxAscii(Port,Filename,Buffer,1024);
          break;
     default:
          DisplayLine(" Must answer Q, S, or R",NULL,0);
          break;
    } /* end switch */
}

/*** Display status line ***/

#define RIGHTMOST 45
int DisplayLine(MsgPtr,UserPtr,UserLength)
char *MsgPtr;
char *UserPtr;
int UserLength;
{static int right = RIGHTMOST;
 int row;
 int col;
 int i;
 char c;
 int MsgLength;
 row = GetRow();
 col = GetCol();
 Position(24,0);
 /* display message */
 MsgLength = strlen(MsgPtr);
 for(i=0;i<MsgLength;i++)
     {AttrWrite(*MsgPtr++,INVERSE);
      Position(24,i+1);
     }
 /* blank rest of menu bar */
 for(i=MsgLength;i<right;i++)
     {AttrWrite(' ',INVERSE);
      Position(24,i+1);
     }
 right = MsgLength;
 /* want response from user ? */
 if(UserPtr!=NULL)
     {right = RIGHTMOST;
      /* input text from user */
      i = 0;
      while(1)
          {Position(24,MsgLength+i);
           c = SioKeyRead();
           if((c=='\r')||(c==ESC)||(MsgLength+i==RIGHTMOST))
               {UserPtr[i] = '\0';
                break;
               }
           if((c==BS)&&(i>0))
               {i--;
                Position(24,MsgLength+i);
                AttrWrite(' ',INVERSE);
               }
           else
               {/* save character & display on status line */
                UserPtr[i++] = c;
                AttrWrite(c,INVERSE);
                if(i==UserLength)
                    {/* user string done */
                     UserPtr[i] = '\0';
                     break;
                    }
               }
          } /* end -- while */
     } /* end -- if */
 Position(row,col);
}

/*** Send character over serial line ***/

int PutChar(Port,ch)
int Port;
char ch;
{int rc;
 rc = SioPutc(Port,ch);
 if(rc<0)
     {printf("COM%d Error: ",1+Port);
      SioError(rc);
      exit(1);
     }
 return(rc);
 }

/*** Receive character from serial line ***/

int GetChar(Port,timeout)
int Port;
int timeout;
{int rc;
 rc = SioGetc(Port,timeout);
 if(rc<-1)
     {printf("COM%d Error: ",1+Port);
      SioError(rc);
      exit(1);
     }
 return(rc);
}

/*** Say error code ***/

int SayError(Port,ptr)
int Port;
char *ptr;
{char temp[81];
 sprintf(temp,"ERROR! COM%d : %s",1+Port,ptr);
 DisplayLine(temp,NULL,0);
 printf("\n*** %s",temp);
 /* cancel remote */
 PutChar(Port,CAN);
 PutChar(Port,CAN);
 PutChar(Port,CAN);
}

/*** Exit gracefully ***/

int MyExit(code,msg)
int code;
char *msg;
{int rc;
 if(code<0) SioError(code);
 /* Assert UART break & sign off */
 SioBrkSig(Port,ASSERT);
 printf("TERMINATING: %s\n",msg);
 SioDelay(8);
 if( (rc = SioDone(Port)) < 0 ) SioError(rc);
 exit(0);
}

char lastXchar;     /* last XON or XOFF sent */
int echo = FALSE;   /* local echo flag */

/*** Say # characters sent or received ***/

int SayCount(Count)
int Count;
{char temp[50];
 /* print message every 200 characters */
 if(Count%200==0)
     {sprintf(temp," %d characters",Count);
      DisplayLine(temp,NULL,0);
     }
}

/*** Say transfer is done ***/

int SayComplete(Count)
int Count;
{char temp[50];
 sprintf(temp,"Transfer Complete: %d characters",Count);
 DisplayLine(temp,NULL,0);
}

/*** Check if user wants to quit ***/

int CheckUser()
{/* does user want to quit ? */
 if(SioKeyPress())
     {SioKeyRead();
      DisplayLine(" Canceled by USER ! ",NULL,0);
      return(TRUE);
     }
 return(FALSE);
}

/*** Send text file ***/

int TxAscii(Port,Filename,Buffer,Length)
int Port;            /* COM port [0..3] */
char Filename[];     /* filename buffer */
char Buffer[];       /* data buffer */
int Length;          /* size of Buffer */
{int i;
 int Code;           /* return code */
 int Handle;         /* file Handle */
 char LastChar;      /* last character sent */
 int TxChars = 0;
 int Count;
 char temp[81];
 /* begin */
 lastXchar =  XON;
 /* Filename is not empty */
 Handle = open(Filename,O_RDONLY|O_BINARY,S_IREAD);
 if(Handle<0)
     {strcpy(temp," Cannot open ");
      strcat(temp,Filename);
      DisplayLine(temp,NULL,0);
      return(FALSE);
     }
 DisplayLine(" ASCII: Starting send",NULL,0);
 while(SioKeyPress()) SioKeyRead();
 /* clear comm port */
 SioRxFlush(Port);
 /* send ascii file ( stop at ^Z ) */
 while(1)
     {/* read next buffer from disk */
      Count = read(Handle,Buffer,Length);
      if(Count==0) break;
      if(Count<0)
          {SayError(Port,"Error on disk read");
           return(FALSE);
          }
      /* send one byte at a time */
      for(i=0;i<Count;i++)
          {/* user want to quit ? */
           if(CheckUser()) return(FALSE);
           /* send byte */
           LastChar = Buffer[i];
           if(echo) printf("%c",LastChar);
           PutChar(Port,LastChar);
           SayCount(++TxChars);
           /* ^Z marks the end of a text file */
           if(LastChar==CTLZ) break;
           /* check for incoming ^Z */
           if(lastXchar==XON)
                {/* check for incoming XON / XOFF */
                 Code = GetChar(Port,0);
                 if(Code>0)
                    {/* is byte a XOFF ? */
                     if((char)Code==XOFF)
                         {/* wait for XON */
                          DisplayLine(" XOFF received",NULL,0);
                          lastXchar = XOFF;
                          /* wait for XON */
                          while(1)
                              {/* user want to quit ? */
                               if(CheckUser()) return(FALSE);
                               Code = GetChar(Port,ONESECOND);
                               if((char)Code==XON)
                                   {DisplayLine(" XON received",NULL,0);
                                    lastXchar = XON;
                                    break;
                                   }
                              } /* end -- while */
                         } /* end -- if(XOFF) */
                    } /* end -- if(Code) */
                } /* end -- if(XON) */

          } /* end -- for(i) */
     } /* end -- while */
 close(Handle);
 if(LastChar!=CTLZ) PutChar(Port,CTLZ);
 SayComplete(TxChars);
 return(TRUE);
} /* end -- TxAcsii */

/*** Receive text file ***/

int RxAscii(Port,Filename,Buffer,Length,RxQueSize)
int Port;            /* COM port [0..3] */
char Filename[];     /* filename buffer */
char Buffer[];       /* data buffer */
int Length;          /* length of data buffer */
int RxQueSize;       /* size of CCL receive buffer */
{int i;
 int Handle;         /* file Handle */
 int Code;           /* return code */
 int lo;             /* receive queue low water mark */
 int hi;             /* receive queue high water mark */
 int Index;          /* buffer index */
 int QueSize;        /* current CCL receive queue size */
 int RxChars = 0;
 int Count;
 char temp[81];
 /* begin */
 lastXchar = XON;
 lo = RxQueSize / 4;
 hi = 3 * lo;
 DisplayLine(" ASCII: Starting receive ",NULL,0);
 while(SioKeyPress()) SioKeyRead();
 /* clear comm port */
 SioRxFlush(Port);
 /* open file passed in Filename[] for write */
 Handle = open(Filename,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,S_IWRITE);
 if(Handle<0)
     {strcpy(temp,"Cannot open ");
      strcat(temp,Filename);
      DisplayLine(temp,NULL,0);
      return(FALSE);
     }
 /* receive text */
 Index = 0;
 while(1)
     {/* user want to quit ? */
      if(CheckUser()) return(FALSE);
      /* check on CCL receive queue size */
      QueSize = SioRxQue(Port);
      if((QueSize>hi)&&(lastXchar==XON))
          {PutChar(Port,XOFF);
           lastXchar = XOFF;
          }
      if((QueSize<lo)&&(lastXchar==XOFF))
          {PutChar(Port,XON);
           lastXchar = XON;
          }
      /* get next byte */
      Code = GetChar(Port,ONESECOND);
      if(Code==-1) continue;
      /* ignore 1st character if it is a 0 */
      if((RxChars==0)&&((char)Code=='\0')) continue;
      /* received a character */
      Buffer[Index++] = (char)Code;
      SayCount(++RxChars);
      /* ^Z marks the end of a text file */
      if((char)Code==CTLZ)
          {/* add 2nd ^Z */
           Buffer[Index++] = CTLZ;
           RxChars++;
           break;
          }
      if(echo) printf("%c",(char)Code);
      if(Index==Length)
          {/* time to write out buffer */
           if(lastXchar==XON)
               {/* send XOFF to sender */
                PutChar(Port,XOFF);
                lastXchar = XOFF;
               }
           /* write disk file */
           Count = write(Handle,Buffer,Index);
           if(Count<0)
               {SayError("Disk read error",NULL);
                SioDelay(18);
                return(FALSE);
               }
           if(lastXchar==XOFF)
               {/* send XON to sender */
                PutChar(Port,XOFF);
                lastXchar = XOFF;
               }
           Index = 0;
          } /* end -- if */
     } /* end -- while */
 /* write any remaining data */
 if(Index>0)
     {Count = write(Handle,Buffer,Index);
      if(Count<0)
          {SayError("Disk read error",NULL);
           SioDelay(18);
           return(FALSE);
          }
     } /* end -- if */
 close(Handle);
 SayComplete(RxChars);
 return(TRUE);
} /* RxAscii */

static BYTE cur_row  = 0;      /* current row */
static BYTE cur_col  = 0;      /* current col */

/*** Write character & attribute ***/

int AttrWrite(ch,attr)
BYTE ch;
BYTE attr;
{union REGS  reg;
 reg.h.ah = 9;
 reg.h.bh = 0;        /* current text page */
 reg.h.al = ch;       /* character */
 reg.h.bl = attr;
 reg.x.cx = 1;
 int86(0x10, &reg, &reg);
}

/*** Position cursor on screen ***/

int Position(row, col)       /* sets current screen position */
BYTE row, col;
{union REGS reg;
 reg.h.ah = 2;
 reg.h.bh = 0;
 reg.h.dh = row;
 reg.h.dl = col;
 int86(0x10, &reg, &reg);
 cur_row = row;
 cur_col = col;
}

/*** Get cursor row ***/

int GetRow()       /* gets current screen cursor row position */
{union  REGS  reg;
 reg.h.ah = 3;
 reg.h.bh = 0;
 int86(0x10, &reg, &reg);
 return(reg.h.dh);
}

/*** Get cursor column ***/

int GetCol()       /* gets current screen cursor col position */
{union  REGS  reg;
 reg.h.ah = 3;
 reg.h.bh = 0;
 int86(0x10, &reg, &reg);
 return(reg.h.dl);
}

/*** Scroll screen ***/

int Scroll(urow, lcol, lrow, rcol, nrows, attr)
unsigned  urow, lcol, lrow, rcol;
int nrows;
int attr;
{union REGS reg;
 reg.h.ah = 6;
 reg.h.ch = urow;
 reg.h.cl = lcol;
 reg.h.al = nrows;
 reg.h.bh = attr;
 reg.h.dh = lrow;
 reg.h.dl = rcol;
 int86(0x10, &reg, &reg);
}